
export {}
declare global {
    namespace TT {

/**
 * When you need a constructor, for example as the argument to a mixin function
 */
type Constructor<I = {}> = new (...args: Array<any>) => I;

/**
 * Extract constructor type of a mixin.
 * Usage: type FooConstructor = MixinConstructorType<typeof Foo>;
 */
type MixinConstructorType<MixinFunction, BaseClass = undefined> = BaseClass extends undefined ? (
    MixinFunction extends (Base: infer B) => infer C ? C : never
) : (
    MixinFunction extends (Base: BaseClass) => infer C ? C : never
);

/**
 * Extract instance type of a mixin
 * Usage: type Foo = MixinType<typeof Foo>;
 */
type MixinType<MixinFunction, BaseClass = undefined> =
    MixinConstructorType<MixinFunction, BaseClass> extends infer R
    ? R extends Constructor
    ? InstanceType<R>
    : never : never;


/*
 * All conditionals follow the pattern:
 *   IfConditional<...Params, TRUE = true, FALSE = false>
 * 
 * The resolve to TRUE if the conditional passes; FALSE otherwise.
 */

/**
 * Conditionally check if property of an interface is optional
 */
type IfKeyIsOptional<T, K extends keyof T, TRUE = true, FALSE = false> = {[K2 in K]: T[K2]} extends {[K2 in K]-?: T[K2]} ? FALSE : TRUE;

// I can't seem to accomplish the same for `readonly`, unfortunately.

/**
 * Conditionally check if any keys exist in both unions.
 */
type IfKeysOverlap<T extends KeyOf, U extends KeyOf, TRUE = true, FALSE = false> = Extract<T, U> | Extract<U, T> extends never ? FALSE : TRUE;

/*
 * Conditional types operating on a union work sorta like Array::map.
 * The result is a new union combining the results of running the conditional against each type in the union.
 * Sometimes, we want something different:
 * We want a conditional to return a single type depending on if *all* members of the input union extend
 * (or do not extend) something.  Sorta like Array::every or Array::some
 */

/** Conditional that tests if all types in a union extend a type */
type IfAllInUnionExtend<Union, Match, ResultIfAllMatch, ResultIfSomeDoNotMatch> = (
    Union extends Match ? never : {__fail_marker: any}
) extends never ? ResultIfAllMatch : ResultIfSomeDoNotMatch;

/** Conditional that tests if all types in a union *do not* extend a type */
type IfNoneInUnionExtend<Union, Match, ResultIfNoneMatch, ResultIfSomeDoMatch> = (
    Union extends Match ? {__fail_marker: any} : never
) extends never ? ResultIfNoneMatch : ResultIfSomeDoMatch;

/** Conditional that tests if at least one type in a union extends a type */
type IfSomeInUnionExtend<Union, Match, ResultIfSomeMatch, ResultIfNoneMatch> = IfNoneInUnionExtend<Union, Match, ResultIfNoneMatch, ResultIfSomeMatch>;
/** Conditional that tests if at least one type in a union *does not* extend a type */
type IfSomeInUnionDoNotExtend<Union, Match, ResultIfSomeDoNotMatch, ResultIfAllMatch> = IfAllInUnionExtend<Union, Match, ResultIfAllMatch, ResultIfSomeDoNotMatch>;


/**
 * When you're too lazy or rushed to use proper type declarations, use TODO
 * This is different from `any` because you *can* apply more appropriate typings;
 * you just choose not to due to time constraints.
 */
type TODO = any;

/**
 * Simple object-backed Dictionary interface.  Slightly less verbose than declaring a subscripting signature.
 */
type Dictionary<T> = Record<string, T>;

type ReadonlyDictionary<T> = Readonly<Record<string, T>>;

/**
 * Backwards-compatible widest type returned by `keyof`
 * TS2.9 expanded `keyof` to include numeric and Symbol properties and index signatures;
 * previously it was limited to `string`
 */
type KeyOf = keyof any;

/**
 * When you want to use `keyof T` but you know the keys are all strings.
 */
type StringKeyOf<T> = Extract<keyof T, string>;

/*
 * Mapped type of T with all K keys omitted.
 */
type Omit<T, K extends KeyOf> = Pick<T, Exclude<keyof T, K>>;

/**
 * Picks only the properties of a certain type.
 * For example, you can filter an interface to include only properties that are numbers.
 * Requires TS 2.8
 */
type PickOfType<Object, T> = {
    [FilteredProp in {[P in keyof Object]: Object[P] extends T ? P : never}[keyof Object]]: Object[FilteredProp];
}

/**
 * Opposite of PickOfType
 * Excludes properties that extend a type; includes all others.
 */
type OmitOfType<Object, T> = {
    [FilteredProp in {[P in keyof Object]: Object[P] extends T ? never : P}[keyof Object]]: Object[FilteredProp];
}

/**
 * When you require something that's invocable.
 * Optionally require a specific return type and/or acceptance of certain argument types,
 * but at that point you should probably write out the function signature directly.
 */
type AnyFunction<Returns = any, A = any, B = any, C = any, D = any, E = any, Rest = any> = (a: A, b: B, c: C, d: D, ...args: Rest[]) => Returns;

/**
 * Sometimes you want to write a class with a readonly property (for the sake of public API)
 * but you (rarely) need to privately mutate that property.
 * Use this trick:
 * 
 *     Writable(this).normallyReadonlyProp = newValue;
 *     // or, if you really want to avoid the function call:
 *     (this as Writable<this>).normallyReadonlyProp = newValue;
 * 
 * Aliased as `Mutable` because that sounds more like the counterpart to Immutable, a term
 * preferred in certain programming circles.
 */
type Writable<T, K extends keyof T = keyof T> = {-readonly [L in K]: T[L]};

// Breaking our rule and putting a value declaration here because it needs to be kept alongside
// the type of the same name

/**
 * Map all keys of a type verbatim.  This preserves all keys / properties, strips
 * invocation signatures, and merges intersection types.
 * 
 * It's common to build an intersection of mapped types.  The resulting `} & {` syntax in
 * tooltips makes them noisier and harder to read.
 * Mapping the result merges all intersected mapped types into a single interface.
 */
type MapKeys<T> = {
    [K in keyof T]: T[K];
}

import { MapKeys } from './core';

// These two are already built-in:
// ReturnType: extracts the return type of a function.
// InstanceType: extracts the instance type of a class.

// ... but what about extracting the arguments of a function?
type ArgumentTypes<Fn extends (...args: any[]) => any> = Fn extends (arg1: infer A, arg2: infer B, arg3: infer C, arg4: infer D, arg5: infer E) => any ? MapKeys<{
    [K in 1]: A
} & {
    [K in 2]: B
} & {
    [K in 3]: C
} & {
    [K in 4]: D
} & {
    [K in 5]: E
}> : any;

// ... or extracting the arguments to a constructor
type ConstructorArgumentTypes<Fn extends new (...args: any[]) => any> = Fn extends new (arg1: infer A, arg2: infer B, arg3: infer C, arg4: infer D, arg5: infer E) => any ? MapKeys<{
    [K in 1]: A
} & {
    [K in 2]: B
} & {
    [K in 3]: C
} & {
    [K in 4]: D
} & {
    [K in 5]: E
}> : any;

// Deprecated single-arg extractors
/** @deprecated Use ArgumentTypes<Fn>[1] instead */
type Argument1Type<Fn extends (arg1: any, ...rest: any[]) => any> = Fn extends (arg1: infer A, ...rest: any[]) => any ? A : any;
/** @deprecated Use ArgumentTypes<Fn>[2] instead */
type Argument2Type<Fn extends (arg1: any, arg2: any, ...rest: any[]) => any> = Fn extends (arg1: any, arg2: infer A, ...rest: any[]) => any ? A : any;
/** @deprecated Use ArgumentTypes<Fn>[3] instead */
type Argument3Type<Fn extends (arg1: any, arg2: any, arg3: any, ...rest: any[]) => any> = Fn extends (arg1: any, arg2: any, arg3: infer A, ...rest: any[]) => any ? A : any;
/** @deprecated Use ArgumentTypes<Fn>[4] instead */
type Argument4Type<Fn extends (arg1: any, arg2: any, arg3: any, arg4: any, ...rest: any[]) => any> = Fn extends (arg1: any, arg2: any, arg3: any, arg4: infer A, ...rest: any[]) => any ? A : any;

/** @deprecated Use ConstructorArgumentTypes<Fn>[1] instead */
type ConstructorArgument1Type<Fn extends new (arg1: any, ...rest: any[]) => any> = Fn extends new (arg1: infer A, ...rest: any[]) => any ? A : any;
/** @deprecated Use ConstructorArgumentTypes<Fn>[2] instead */
type ConstructorArgument2Type<Fn extends new (arg1: any, arg2: any, ...rest: any[]) => any> = Fn extends new (arg1: any, arg2: infer A, ...rest: any[]) => any ? A : any;
/** @deprecated Use ConstructorArgumentTypes<Fn>[3] instead */
type ConstructorArgument3Type<Fn extends new (arg1: any, arg2: any, arg3: any, ...rest: any[]) => any> = Fn extends new (arg1: any, arg2: any, arg3: infer A, ...rest: any[]) => any ? A : any;
/** @deprecated Use ConstructorArgumentTypes<Fn>[4] instead */
type ConstructorArgument4Type<Fn extends new (arg1: any, arg2: any, arg3: any, arg4: any, ...rest: any[]) => any> = Fn extends new (arg1: any, arg2: any, arg3: any, arg4: infer A, ...rest: any[]) => any ? A : any;


/**
 * Design-time equivalent of `Object.assign`
 * Pros:
 *   overwrites properties correctly
 * Cons:
 *   strips function and construct signatures from A
 *   strips JSDoc from all props
 */
type Overlay<A, B> = IfKeysOverlap<keyof A, keyof B, Pick<A, Exclude<keyof A, keyof B>> & {[P in keyof B]: IfKeyIsOptional<B, P, B[P] | (P extends keyof A ? A[P] : never), B[P]>}, A & B>;


    }
}
